/**HuffFM.cpp*********************************************************************************************
 *          Copyright 2004, V. Makinen                                                                   *
 *                                                                                                       *
 * This file contains an implementation of a succinct full-text index: HuffFM-index.           	      	 * 
 * For updates, see: http://www.cs.helsinki.fi/u/vmakinen/software                                       *
 * For more information, see [MN04].        	      	                                                 *                     
 *                                                                                                       *
 *                                                                                                       *   
 * This software may be used freely for any non-commercial purpose. However,                             *
 * when distributed, the original source must be clearly stated, and, when the                           *
 * source code is distributed, the copyright notice must be retained and any                             *
 * alterations in the code must be clearly marked. No warranty is given regarding                        *
 * the quality of this software.                                                                         *
 *                                                                                                       *
 * [GMN04] Szymon Grabowski, Veli Mkinen, and Gonzalo Navarro. First Huffman, then Burrows-Wheeler -    *    
 *        A Simple Alphabet-Independent FM-Index. To appear in SPIRE 2004.              	      	 *
 *                                                                                                       *
 *********************************************************************************************************/


//---------------------------------------------------------------------------
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/times.h>
#include <assert.h>
#include <string.h>
#include "qsufsort.c"
#include "maketable.cpp"

//---------------------------------------------------------------------------

#define W 32 // computer word size
#define Wminusone 31  // W-1
#define WW 64  // 2*W

unsigned n,m; // text length, pattern length
bool print=true;
int report=1;  // 0=existence, 1=report, 2=count, X>2 = print X characters after each pattern occurrence
unsigned bitsn=0; // bits needed to represent n
unsigned samplerate=64; // every samplerare^th suffix is stored as such

char* text;

class TSA;
class TBitRank;
class THuffFMindex;
char *BWT(char *text, unsigned n);
TSA *CreateSuffixArray(char *text, unsigned n);

inline void SetField(unsigned *A,register unsigned len, register unsigned index,register unsigned x) {
   unsigned i=index*len/W,
            j=index*len-i*W;
   unsigned mask = (j+len < W ? ~0u << j+len : 0)
                   | (W-j < W ? ~0u >> (W-j) : 0);
   A[i] = (A[i] & mask) | x << j;
   if (j+len>W) {
      mask = ((~0u) << (len+j-W));  
      A[i+1] = (A[i+1] & mask)| x >> (W-j);
   } 
}; 

inline void SetVariableField(unsigned *A,register unsigned len, register unsigned index,register unsigned x) {
   unsigned i=index/W, 
            j=index-i*W;
   unsigned mask = (j+len < W ? ~0u << j+len : 0) 
                   | (W-j < W ? ~0u >> (W-j) : 0);
   A[i] = (A[i] & mask) | x << j;
   if (j+len>W) {
      mask = ((~0u) << (len+j-W));  
      A[i+1] = (A[i+1] & mask)| x >> (W-j);
   } 
}; 

inline unsigned GetField(unsigned *A, register  unsigned len, register unsigned index) {
   register unsigned i=index*len/W, 
            j=index*len-W*i,
            result;                
   if (j+len <= W)
      result = (A[i] << (W-j-len)) >> (W-len);
   else {
      result = A[i] >> j;
      result = result | (A[i+1] << (WW-j-len)) >> (W-len);
   }
   return result;
};

inline unsigned bitsneeded(unsigned x) {return (unsigned)ceil(log((double)x)/log(2.0))+1;};

class TSA {
public:
    unsigned posbits;
    unsigned *pos;
    unsigned *B;    
    unsigned n;
public:
    void CreatePosArray(unsigned bits, unsigned n) 
      {this->n = n; posbits = bits; pos = new unsigned[n*posbits/W+2];
       CreateBooleanArray(n);};
    void CreateBooleanArray(unsigned n)
      {B = new unsigned[n/W+2];};   
    void DeletePosArray() {delete pos;};
    void DeleteBooleanArray() {delete B;};        
    void SetPos(unsigned index, unsigned x)
      {SetField(pos,posbits,index,x);};
    void SetB(unsigned index, unsigned x)
      {SetField(B,1,index,x);};
    unsigned GetPos(unsigned index)
      {return GetField(pos,posbits,index);};        
    bool GetB(unsigned index)
      {return GetField(B,1,index);};        
    ~TSA() {DeletePosArray(); DeleteBooleanArray();};
};

TSA *CreateSuffixArray(char *text, unsigned n)
{
   TSA *SA;
   int *x, *p;
   p=(int*)malloc((n+1)*sizeof *p);
   x=(int*)malloc((n+1)*sizeof *x);
   unsigned i;
   for (i=0; i<n; i++)
      x[i]=(int)text[i]+128;
   x[n]=0;
   suffixsort(x,p,n,256,0);
   free(x);
   unsigned k=0;
   while ((n<<k)>>(W-1) == 0) k++;
   bitsn = W-k;
   SA = new TSA();
   SA->CreatePosArray(bitsn,n);
   for (i=0; i<n; i++) {
      SA->SetPos(i,p[i+1]);
      SA->SetB(i,0);
   }
   free(p);
   return SA;
}


class TBitRank {
/* Supports rank(i)=(sum_{0<=i'<=i} B[i]) operation in constant time on bit-array B */
/* Takes less than n/(log n)+2n(loglog n)/(log n)+sqrt(n)(log n)(loglog n)+n bits */
/* That is, it takes n+o(n) bits (in practice about 1.6n bits) */
/* Also supports operation select(x)=i, where rank(i)=x. */
/* Select is implemented making a binary search over the rank-structure,
/* so it takes O(log(n)) time. It is possible to implement O(1)-time select,
/* using similar techniques as with rank.*/
/* See more details in David Clark's PhD thesis. */
private:
   unsigned *firstrank; // rank of each log^2 n:th bit
   unsigned firstbits; // field length of firstrank (log n)
   unsigned firstbitspow2; // =firstbits*firstbits for efficiency
   unsigned firstn; // length of firstrank (n/log^2 n)/W
   unsigned *secondrank; // rank of each log n:th bit inside log^2 n size field
   unsigned secondbits; // field length of firstrank (2 log log n)
   unsigned secondn; // length of secondrank (2 n log log n/log n)/W
   unsigned *thirdrank; // ranks of all sqrt(n) fields of length (log n)/2
   unsigned thirdbits;  // field length of thirdrank log((log n)/2)
   unsigned thirdn;  // length of thirdrank sqrt(n)(log((log n)/2)(log n)/2)/W
   bool owndata;
   unsigned *data; // stored version of B
   unsigned databits;
public:
   unsigned n; // length of bit-array B
private:
   void MakeRankDictionary(bool *B) {
      unsigned i,j;
      unsigned sum=0;
      bool copydata = false;
	  if (B == NULL) {
		 copydata = true;
	     B = new bool[n];
	     for (i=0;i<n;i++)
	        B[i] = IsBitSet(i);
      }      
      /* First-level dictionary */  
      firstbits = bitsneeded(n);
      firstn = (unsigned)floor((floor(n/(1.0*firstbits))+1)*1.0/W+1)+1;
      firstrank = new unsigned[firstn];
      firstbitspow2 = firstbits*firstbits;
      j=0;
      for (i=0;i*firstbitspow2<n+firstbitspow2;i++) {
         for (;j<=i*firstbitspow2 && j<n;j++)
            sum += B[j];
         SetField(firstrank,firstbits,i,sum);
      }
      /* Second-level dictionary */
      secondbits = bitsneeded(firstbitspow2);
      secondn = (unsigned)floor((floor(secondbits*n/(1.0*firstbits))+1)*1.0/W+1)+1+firstbits*secondbits/W+1;
      secondrank = new unsigned[secondn];
      sum = 0;
      j = 0;
      for (i=0;i*firstbits<n+firstbitspow2;i++) {
         if ((i*firstbits % (firstbitspow2))==0) {
            j = i*firstbits+1;
            sum = 0;
         }
         for (;j<=i*firstbits && j<n;j++)
            sum += B[j];
         SetField(secondrank,secondbits,i,sum);
      }
        /* Third-level dictionary */
      unsigned lognpertwo = (unsigned)floor(firstbits/2.0)+1;
       unsigned x;
      thirdbits = bitsneeded(lognpertwo);
      unsigned thirdlength = (1u << lognpertwo);
      thirdn = (unsigned)floor(1.0*thirdlength*lognpertwo*thirdbits/W)+1;
      thirdrank = new unsigned[thirdn];
       for (i=0;i<thirdlength;i++) {
          sum = 0;
          for (j=0;j<lognpertwo;j++) {
            if (i & (1u<<j)) sum++;
            SetField(thirdrank,thirdbits,lognpertwo*i+j,sum);
          }
        }
       databits = lognpertwo;
       if (copydata) {delete B; return;};
       // convert B into unsigned[]
       data = new unsigned[n/W+1];
       for (i=0;i<n;i++) {
          if (i % W == 0) data[i/W]=0;
          data[i/W] = data[i/W] | (unsigned)B[i] << (i % W);
       }
       //caller can delete B
   };
   inline unsigned rank1(unsigned i) {
      return GetField(firstrank,firstbits,i/(firstbitspow2));
   };
   inline unsigned rank2(unsigned i) {
     return GetField(secondrank,secondbits,i/firstbits);
   };
   inline unsigned rank3(unsigned i) {
      unsigned j = GetField(data,firstbits,i/firstbits);
      unsigned k = i % firstbits;
      bool isbitset = 1u & j;
      unsigned rankthree;
      if (k < databits) {
        j = (j << (W-databits)) >> (W-databits);
         rankthree = GetField(thirdrank,thirdbits,j*databits+k);
      }
      else {
         unsigned jj = j >> databits;
         j = (j << (W-databits)) >> (W-databits);
         rankthree = GetField(thirdrank,thirdbits,(j+1)*databits-1);
         k -= databits;
         rankthree += GetField(thirdrank,thirdbits,jj*databits+k);
      }
      return isbitset?rankthree-1:rankthree;
   };
public:
   TBitRank(bool *B, unsigned n) {this->n=n; owndata = true; MakeRankDictionary(B);};
   TBitRank(unsigned *data, unsigned n, bool owndata) {this->n=n; this->data = data; this->owndata = owndata; MakeRankDictionary(NULL);};
   unsigned SpaceRequirementInBits() {
      return firstn*W+secondn*W+thirdn*W+(owndata?n:0);
   };
   inline unsigned rank(unsigned i) { // returs sum_{0<=i'<=i} B[i]
      return rank1(i)+rank2(i)+rank3(i);
   };
   inline bool IsBitSet(unsigned i) {
      return (1u << (i % W)) & data[i/W];
   };
   unsigned selectBasic(unsigned x) { // returns i such that x=rank(i) && rank(i-1)<x
      unsigned l=0, r=n;
      unsigned mid=(l+r)/2;
      unsigned rankmid = rank(mid);
      while (l+1<r) {
        if (x>rankmid) {
          l=mid;
          mid = (l+r)/2;
        } else {
          r = mid;
          mid = (l+r)/2;
        }
        rankmid = rank(mid);
      }
      if (rankmid == x) return l;
      else return r;
   };
   inline unsigned select(unsigned x) { // returns i such that x=rank(i) && rank(i-1)<x
      // first binary search over first level rank structure
      // then binary search over second level rank structure
      // then binary search over third level rank structure
      // this is 6-7 times faster than selectBasic
      unsigned rankr;
      unsigned l=0, r=n/(firstbitspow2)+1;
      unsigned mid=(l+r)/2;
      unsigned rankmid = GetField(firstrank,firstbits,mid);
      while (l<mid) {
        if (x>rankmid)
          l = mid;
         else
          r = mid;
       mid = (l+r)/2;
        rankmid = GetField(firstrank,firstbits,mid);
      }
      unsigned left;
      unsigned rankleft;
      left=l*firstbits;
      rankleft = rankmid;

      // now correct value is between left*firstbits+1...(left+firstbits)*firstbits
      l=0; r = firstbits;
      mid=(l+r)/2;
      rankmid = GetField(secondrank,secondbits,left+mid);
      while (l<mid) {
        if (x>rankleft+rankmid)
          l = mid;
        else
          r = mid;
        mid = (l+r)/2;
        rankmid = GetField(secondrank,secondbits,left+mid);
      }
      left += l;
      rankleft += GetField(secondrank,secondbits,left);

      // now correct value is between left*firstbits+1...(left+1)*firstbits-1
      unsigned j = GetField(data,firstbits,left);
      unsigned jj = j >> databits; //sedondhalf
      j = (j << (W-databits)) >> (W-databits);
      rankleft -= 1u & j;
      rankmid = GetField(thirdrank,thirdbits,(j+1)*databits-1);
      r = databits;
      if (rankleft+rankmid < x) {
         j = jj;
         left = left*firstbits+databits;
         rankleft += rankmid;
         r = firstbits-databits;
      }
      else left = left*firstbits;
      l=0;
      mid=(l+r)/2;
      rankmid = GetField(thirdrank,thirdbits,j*databits+mid);
      while (l<mid) {
        if (x>rankleft+rankmid)
          l = mid;
        else
          r = mid;
        mid = (l+r)/2;
        rankmid = GetField(thirdrank,thirdbits,j*databits+mid);
      }
      if (rankleft+rankmid == x) return left+l;
      else return left+r;
   }
   inline unsigned prev(unsigned start) {
      // returns the position of the previous 1 bit before and including start.
      // tuned to 32 bit machine

      unsigned i = start >> 5;
      int offset = (start % W);
      unsigned answer = start;
      unsigned int val = data[i] << (Wminusone-offset);

      if (!val) { val = data[--i]; answer -= 1+offset; }

      while (!val) { val = data[--i]; answer -= W; }

      if (!(val & 0xFFFF0000)) { val <<= 16; answer -= 16; }
      if (!(val & 0xFF000000)) { val <<= 8; answer -= 8; }

      while (!(val & 0x80000000)) { val <<= 1; answer--; }
      return answer;
      /*
      // simpler, but 3 times slower:
      while (!IsBitSet(start)) --start;
      return start;
      */

   }
   inline unsigned next(unsigned start) {
	  // this could be done more efficiently like prev()
      while (!IsBitSet(start)) ++start;
      return start;
   }
   void print() {
      unsigned i;
      for (i=0;i<n;i++)
         if (IsBitSet(i))
	    printf("%c",'1');
	 else
	    printf("%c",'0');
      printf("\n");
   }
   void Test() {
      unsigned i, iprev=0;
      for (i=0;i<n;i++) {
        if (rank(i)>rank(iprev)) iprev = i;
         if (rank(i)>0 && select(rank(i))!=iprev)
            printf("select(%d)=%d!=%d\n",rank(i),select(rank(i)),iprev);
       }
   };
   ~TBitRank() {
      delete firstrank;
      delete secondrank;
      delete thirdrank;
      if (owndata)
         delete data;
   };
};

/* The code below is not used yet, since it is only needed for reporting queries */
/* Anyway, the above brute-force prev() algorithm of BitRank is usually faster */

class TBitSelectPrev {
/* Supports prev(i)=(max_{0<=i'<=i, B[i']=1} i') operation in constant time on bit-array B */
/* Takes less than n/(log n)+2n(loglog n)/(log n)+sqrt(n)(log n)(loglog n)+n bits */
/* That is, it takes n+o(n) bits (in practice about 1.6n bits) */
/* This structure is due to Szymon Grabowski (personal communication) */
private:
   unsigned *firstprev; // selectprev of each log^2 n:th bit
   unsigned firstbits; // field length of firstrank (log n)
   unsigned firstbitspow2; // =firstbits*firstbits for efficiency
   unsigned firstn; // length of firsprev (n/log^2 n)/W
   unsigned *secondprev; // selectprev of each log n:th bit inside log^2 n size field
   unsigned secondbits; // field length of secondprev (2 log log n)
   unsigned secondn; // length of secondprev (2 n log log n/log n)/W
   unsigned *thirdprev; // selectprev of all sqrt(n) fields of length (log n)/2
   unsigned thirdbits;  // field length of thirdprev log((log n)/2)
   unsigned thirdn;  // length of thirdprev sqrt(n)(log((log n)/2)(log n)/2)/W
   bool owndata; // do we own data
   unsigned *data; // stored version of B
   unsigned databits;
public:
   unsigned n; // length of bit-array B
private:
   void MakeSelectPrevDictionary(bool *B) {
      unsigned i,j;
      unsigned *prevc = new unsigned[n];
      j = n+1;
      if (B!=NULL) {
         for (i=0;i<n;i++) { 
            if (B[i]) j = i;
            prevc[i] = j;
         }
      }
      else {
         for (i=0;i<n;i++) {
            if (IsBitSet(i)) j = i;
            prevc[i] = j;
         }
      }

      /* First-level dictionary */
      firstbits = bitsneeded(n);
      firstn = (unsigned)floor((floor(n/(1.0*firstbits))+1)*1.0/W+1)+1;
      firstprev = new unsigned[firstn];
      j=0;
      firstbitspow2 = firstbits*firstbits;
      for (i=0;i*firstbitspow2<n+firstbitspow2;i++) 
         SetField(firstprev,firstbits,i,prevc[i*firstbitspow2]); 
      /* Second-level dictionary */
      secondbits = bitsneeded(firstbitspow2);
      secondn = (unsigned)floor((floor(secondbits*n/(1.0*firstbits))+1)*1.0/W+1)+1+firstbits*secondbits/W+1;
      secondprev = new unsigned[secondn];
      j = 0;
      for (i=0;i*firstbits<n+firstbitspow2;i++) {
         if ((i*firstbits % (firstbitspow2))==0)
            j = i*firstbits;
         if (prevc[i*firstbits]<j || prevc[i*firstbits]>i*firstbits) 
            // can only afford to store differences not greater than log^2 n 
            SetField(secondprev,secondbits,i,i*firstbits-j); 
         else 
            SetField(secondprev,secondbits,i,i*firstbits-prevc[i*firstbits]); 
      }    
      /* Third-level dictionary */ 
      unsigned lognpertwo = (unsigned)floor(firstbits/2.0)+1; 
      unsigned x,k;
      thirdbits = bitsneeded(lognpertwo);
      unsigned thirdlength = (1u << lognpertwo); 
      thirdn = (unsigned)floor(1.0*thirdlength*lognpertwo*thirdbits/W)+1;
      thirdprev = new unsigned[thirdn];
      for (i=0;i<thirdlength;i++) {
          k = lognpertwo;
          for (j=0;j<lognpertwo;j++) {
            if (i & (1u<<j)) k=j;
            SetField(thirdprev,thirdbits,lognpertwo*i+j,k);    
          }              
        }         
       delete prevc;               
       databits = lognpertwo;
       if (B == NULL) return;
       // convert B into unsigned[]
       data = new unsigned[n/W+1];
       for (i=0;i<n;i++) {
          if (i % W == 0) data[i/W]=0;
          data[i/W] = data[i/W] | (unsigned)B[i] << (i % W);
       }       
       //caller can delete B   
   };           
   inline unsigned prev1(unsigned i) {
      return GetField(firstprev,firstbits,i/(firstbitspow2));      
   };      
   inline unsigned prev2(unsigned i) {
     return (i/firstbits)*firstbits-GetField(secondprev,secondbits,i/firstbits);
   };
   inline unsigned prev3(unsigned i) {
      unsigned j = GetField(data,firstbits,i/firstbits);
      unsigned k = i % firstbits;
      unsigned prevthree;
      if (k < databits) {
         j = (j << (W-databits)) >> (W-databits);
         prevthree = GetField(thirdprev,thirdbits,j*databits+k);
         if (prevthree == databits) return i+1; // not found using prev3
         else return i+prevthree-k;  // prev found    
      }
      else {         
         unsigned jj = j >> databits;
         k -= databits;         
	     prevthree = GetField(thirdprev,thirdbits,jj*databits+k);
         if (prevthree != databits) return i+prevthree-k; // prev found	                                    
         // prev not yet found
         j = (j << (W-databits)) >> (W-databits);         
         prevthree = GetField(thirdprev,thirdbits,(j+1)*databits-1);
         if (prevthree != databits) return i+prevthree-k-databits; // prev found
         else return i+1; // not found using prev3     
      }
   };
public:   
   TBitSelectPrev(bool *B, unsigned n) {this->n=n; owndata = true; MakeSelectPrevDictionary(B);};
   TBitSelectPrev(unsigned *data, unsigned n, bool owndata) {this->data=data; this->n=n; this->owndata = owndata; MakeSelectPrevDictionary(NULL);};
   unsigned SpaceRequirementInBits() {
      return firstn*W+secondn*W+thirdn*W+(owndata?n:0);
   }; 
   inline unsigned prev(unsigned i) { // returs max_{j<=i, B[j]=1} j
      unsigned result = prev3(i);
      if (result <= i) return result;
      result = prev2(i);
      if (result % (firstbitspow2) == 0) 
         return prev1(result);
      else 
         return result;               
   };
   inline bool IsBitSet(unsigned i) {
      return (1u << (i % W)) & data[i/W];
   };      
   inline unsigned prevbf(unsigned start) {
      // returns the position of the previous 1 bit before and including start.
      // tuned to 32 bit machine
      
      unsigned i = start >> 5;
      int offset = (start % W);
      unsigned answer = start;
      unsigned int val = data[i] << (Wminusone-offset);

      if (!val) { val = data[--i]; answer -= 1+offset; }

      while (!val) { val = data[--i]; answer -= W; }

      if (!(val & 0xFFFF0000)) { val <<= 16; answer -= 16; }
      if (!(val & 0xFF000000)) { val <<= 8; answer -= 8; }

      while (!(val & 0x80000000)) { val <<= 1; answer--; }
      return answer;
      /*
      // simpler, but 3 times slower:
      while (!IsBitSet(start)) --start;
      return start;
      */

   }
   void Test() {
      unsigned i,j,k;
      j=0;
      unsigned *prevc = new unsigned[n];
      for (i=0;i<n; i++) {
         if (IsBitSet(i)) j=i;
         prevc[i] = j;
      }
      for (i=0;i<n;i++) {
	     j = prevc[i];
	     k = prev(i);
	     if (j!=k)
	        printf("Error: prev(%d)=%d<>%d=prev[%d]\n",i,k,j,i);
      }
      delete prevc;
   }
   void Testbf() {
      unsigned i,j,k;
      j=0;
      unsigned *prevc = new unsigned[n];
      for (i=0;i<n; i++) {
         if (IsBitSet(i)) j=i;
         prevc[i] = j;
      }
      for (i=0;i<n;i++) {
	     j = prevc[i];
	     k = prevbf(i);
	     if (j!=k)
	        printf("Error: prev(%d)=%d<>%d=prev[%d]\n",i,k,j,i);
      }
      delete prevc;
   }
   inline unsigned next(unsigned start) {
      // simpler, but 3 times slower:
      while (!IsBitSet(start)) ++start;
      return start;
   }
   ~TBitSelectPrev() {
      delete firstprev;
      delete secondprev;
      delete thirdprev;
      if (owndata)
         delete data;
   };
};



/**********************************************************************************************************
 * FM-index implementation using bit ranking and Huffman-compression: HuffFM-index                        *
 **********************************************************************************************************/

unsigned *BWT(char *text, unsigned len, unsigned *blocksdata, unsigned *lastsuffixpos, unsigned n, unsigned **markedsamples, unsigned **suffixes) {
   TSA *SA = CreateSuffixArray(text,len);

   unsigned *s = new unsigned[len/W+1];
   bool *suffix = new bool[len];
   *markedsamples = new unsigned[n/W+1];
   *suffixes = new unsigned[n/samplerate+1];
   unsigned i;

   for (i=0;i<n/W+1;i++)
      (*markedsamples)[i]=0u;
   for (i=0;i<n/samplerate+1;i++)
      (*suffixes)[i]=0u;

   // BWT
   for (i=0;i<len;i++)
      if (SA->GetPos(i)==0) {SetField(s,1,i,0); *lastsuffixpos = i;}
      else
         SetField(s,1,i,text[SA->GetPos(i)-1]=='1');
   unsigned *sampleblocksdata  = new unsigned[len/W+1];
   unsigned count=0;
   for (i=0;i<len;i++)
      if (GetField(blocksdata,1,i)==1 && (count % samplerate==0)) {
         SetField(sampleblocksdata,1,i,1);
	 count++;
      }
      else {
         SetField(sampleblocksdata,1,i,0);
	 if (GetField(blocksdata,1,i)==1) count++;
      }
   count = 0;
   for (i=0;i<len;i++) {
      suffix[i] = GetField(blocksdata,1,SA->GetPos(i));
      if (GetField(sampleblocksdata,1,SA->GetPos(i))) {
         (*suffixes)[count++] = SA->GetPos(i); // first, store a pointer
      }
   }
   // remap sampledata
   TBitRank *srank = new TBitRank(blocksdata,len,false);
   for (i=0;i<n/samplerate+1;i++)
      (*suffixes)[i] = srank->rank((*suffixes)[i]);
   delete srank;
   count=0;
   for (i=0;i<len;i++)
      if (suffix[i] && GetField(sampleblocksdata,1,SA->GetPos(i)))
         SetField(*markedsamples,1,count++,1);
      else if (suffix[i])
         SetField(*markedsamples,1,count++,0);

   // remap blocksdata
   for (i=0;i<len;i++)
      SetField(blocksdata,1,i,suffix[i]);

   delete suffix;
   delete SA;
   delete sampleblocksdata;
   return s;
}


class THuffFMindex {
public:
    unsigned n; // original text length (in bytes)
    unsigned len; // huffman compressed length (in bits)
    unsigned zeros; // number of zeros in huffcompressed text
    unsigned lastsuffixpos; // SA pos of last suffix
    unsigned lastposiszero; // last bit of huffman compressed text
    TBitRank *bwthuff; // huff+bwt encoded text + rank structures
    TBitRank *blocks; // suffixes corresponding to codes marked as 1-bits
    TBitRank *sampled; // marking of sampled suffixes
    unsigned *suffixes; // occurrence positions of sampled suffixes
    TCodeEntry *codetable; // huffman codes for chars 0...255
public:
    THuffFMindex(char *text, unsigned n) {
       printf("Text size is %d B\n",n);
       // First huffman compress the text
       // into *huffcode*.
       // Also keep in *blocks* the code boundaries.
       // Then construct the bwt of *huffcode* into *bwthuff* and
       // change *blocks* to refer code-starting suffixes of *bwthuff*
       codetable = makecodetable(text,n);
       unsigned i,j;
       // length of the code
       len = 0;
       for (i=0;i<256;i++)
          len += codetable[i].bits*codetable[i].count;
       unsigned *blocksdata = new unsigned[len/W+1];
       for (i=0;i<len/W+1;i++)
          blocksdata[i] = 0u;
       // coding
       unsigned pos=0;
       char *huffcodeunary = new char[len];
       for (i=0;i<n;i++) {
          for (j=0;j<codetable[(int)text[i]+128].bits;j++) {
             if (codetable[(int)text[i]+128].code & (1u << j))
                huffcodeunary[pos+j] = '1';
             else
                huffcodeunary[pos+j] = '0';
	  }
          SetField(blocksdata,1,pos,1);
          pos += codetable[(int)text[i]+128].bits;
       }

       // caller can delete text
       if (pos != len) printf("error: pos=%d, len=%d\n",pos,len);
       huffcodeunary[len] = 0;
       unsigned *markedpositions;
       lastposiszero = (huffcodeunary[len-1]=='0');
       unsigned *bwthuffdata = BWT(huffcodeunary,len,blocksdata,&lastsuffixpos,n,&markedpositions,&suffixes);
       delete huffcodeunary;
       bwthuff = new TBitRank(bwthuffdata,len,true);
       blocks = new TBitRank(blocksdata,len,true);
       sampled = new TBitRank(markedpositions,n,true);
       zeros = len - bwthuff->rank(len-1);
       printf("Length of bwt(huff(text))=%d bits\n",len);
       // only taking into account the codes from codetable, other information can be removed
       printf("HuffFMindex takes %d B\n",256*W/8+W*n/(samplerate*8)+bwthuff->SpaceRequirementInBits()/8 +
               blocks->SpaceRequirementInBits()/8+sampled->SpaceRequirementInBits()/8);
       fflush(stdout);
    };
    int Search(char *pattern, unsigned m) {
       // use the FM-search replacing function Occ(c,1,i) with bwthuff->rank(c,i), and blocks->rank(c,i)
       // first huffman compress pattern

       // upper bound to the length; this could be done better but more slowly
       char *huffpat = new char[m*32];
       int i, j, plen=0, occ;
       for (i=0;i<m;i++) {
          if (codetable[(int)pattern[i]+128].bits == 0) return 0; // symbol not in text
          for (j=0;j<codetable[(int)pattern[i]+128].bits;j++)
             if (codetable[(int)pattern[i]+128].code & (1u << j))
                huffpat[plen+j] = '1';
             else
                huffpat[plen+j] = '0';
          plen += codetable[(int)pattern[i]+128].bits;
       }
       i = plen-1;
       int sp, ep;
       if (lastposiszero) {
	  // A[0] is the last suffix, let's prevent traversing from hufftext[0] to hufftext[len-1]
          sp = 1; ep = len-1;
          while ((sp<=ep) && (i >=0)) {
             if (huffpat[i]=='0') {
                sp = sp - bwthuff->rank(sp-1)+1-(sp-1>=lastsuffixpos);
                ep = ep - bwthuff->rank(ep)+1-(ep>=lastsuffixpos);
             }
             else {
                sp = zeros + bwthuff->rank(sp-1);
                ep = zeros + bwthuff->rank(ep)-1;
             }
             if (sp==0) ++sp;
             --i;
          }
       }
       else {
          // A[zeros] is the last suffix, let's prevent traversing from hufftext[0] to hufftext[len-1]
	  sp = 0; ep = len-1;
          while ((sp<=ep) && (i >=0)) {
             if (huffpat[i]=='0') {
                sp = sp>0?(sp - bwthuff->rank(sp-1)+1-(sp-1>=lastsuffixpos)):0;
                ep = ep - bwthuff->rank(ep)+1-(ep>=lastsuffixpos);
             }
             else {
                sp = zeros + (sp>0?(bwthuff->rank(sp-1)):0);
                ep = zeros + bwthuff->rank(ep);
             }
             if (sp==zeros) ++sp;
             if (ep==zeros) --ep;
             --i;
          }
       }
       delete huffpat;
       unsigned dist;
       if (ep<sp) return 0;
       if (report==1) {
          i = blocks->prev(ep);
	  while (i>=sp) {
	     j=i; dist=0;
             while (!blocks->IsBitSet(j) || !sampled->IsBitSet(blocks->rank(j)-1)) {
	        // LF mapping
	        j = bwthuff->IsBitSet(j)?zeros+bwthuff->rank(j)-1:j-bwthuff->rank(j)+1-(j>=lastsuffixpos);
		if (blocks->IsBitSet(j)) ++dist;
	     }
	     if (print) {
	        printf("Match at %d\n",suffixes[sampled->rank(blocks->rank(j)-1)-1]+dist-1);
		fflush(stdout);
	     }
	     i = blocks->prev(i-1);
	  }
	  return blocks->rank(ep)-(sp>0?blocks->rank(sp-1):0);
       }
       else return blocks->rank(ep)-(sp>0?blocks->rank(sp-1):0);
    };
    ~THuffFMindex() {
       delete bwthuff;
       delete blocks;
       delete[] codetable;
    };
};



