/**RLFM.cpp***********************************************************************************************
 *          Copyright 2004, V. Makinen                                                                   *
 *                                                                                                       *
 * This file contains an implementation of a succinct full-text index: RLFM-index.           	      	 * 
 * For updates, see: http://www.cs.helsinki.fi/u/vmakinen/software                                       *
 * For more information, see [MN04].        	      	                                                 *                     
 *                                                                                                       *
 *                                                                                                       *   
 * This software may be used freely for any non-commercial purpose. However,                             *
 * when distributed, the original source must be clearly stated, and, when the                           *
 * source code is distributed, the copyright notice must be retained and any                             *
 * alterations in the code must be clearly marked. No warranty is given regarding                        *
 * the quality of this software.                                                                         *
 *                                                                                                       *
 * [MN04] Veli Mkinen and Gonzalo Navarro. New search algorithms and space/time tardeoffs               *    
 *        for succinct suffix arrays                                                                     *
 *        Technical report C-2004-20, Dept. CS, Univ. Helsinki, April 2004.    	      	             	 *
 *                                                                                                       *
 *********************************************************************************************************/



//---------------------------------------------------------------------------
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/times.h>
#include <assert.h>
#include <string.h>
#include "qsufsort.c"
//---------------------------------------------------------------------------

#define W 32 // computer word size
#define Wminusone 31  // W-1
#define WW 64  // 2*W

unsigned n,m; // text length, pattern length
bool print=true; 
int report=1;  // 0=existence, 1=report, 2=count, X>2 = print X characters after each pattern occurrence
unsigned bitsn=0; // bits needed to represent n

char* text; 

class TSA;
class TBitRank;
class TAlphabetRank;
class TFMindexRL;
char *BWT(char *text, unsigned n);
void TestAlphabetRank();
TSA *CreateSuffixArray(char *text, unsigned n);

inline void SetField(unsigned *A,register unsigned len, register unsigned index,register  unsigned x) {
   unsigned i=index*len/W, 
            j=index*len-i*W;
   unsigned mask = (j+len < W ? ~0u << j+len : 0) 
                   | (W-j < W ? ~0u >> (W-j) : 0);
   A[i] = (A[i] & mask) | x << j;
   if (j+len>W) {
      mask = ((~0u) << (len+j-W));  
      A[i+1] = (A[i+1] & mask)| x >> (W-j);
   } 
}; 

inline unsigned GetField(unsigned *A, register  unsigned len, register unsigned index) {
   register unsigned i=index*len/W, 
            j=index*len-W*i,
            result;                
   if (j+len <= W)
      result = (A[i] << (W-j-len)) >> (W-len);
   else {
      result = A[i] >> j;
      result = result | (A[i+1] << (WW-j-len)) >> (W-len);
   }
   return result;
};

inline unsigned bitsneeded(unsigned x) {return (unsigned)ceil(log((double)x)/log(2.0))+1;};

class TSA {
public:
    unsigned posbits;
    unsigned *pos;
    unsigned *B;    
    unsigned n;
public:
    void CreatePosArray(unsigned bits, unsigned n) 
      {this->n = n; posbits = bits; pos = new unsigned[n*posbits/W+2];
       CreateBooleanArray(n);};
    void CreateBooleanArray(unsigned n)
      {B = new unsigned[n/W+2];};   
    void DeletePosArray() {delete pos;};
    void DeleteBooleanArray() {delete B;};        
    void SetPos(unsigned index, unsigned x)
      {SetField(pos,posbits,index,x);};
    void SetB(unsigned index, unsigned x)
      {SetField(B,1,index,x);};
    unsigned GetPos(unsigned index)
      {return GetField(pos,posbits,index);};        
    bool GetB(unsigned index)
      {return GetField(B,1,index);};        
    ~TSA() {DeletePosArray(); DeleteBooleanArray();};
};

TSA *CreateSuffixArray(char *text, unsigned n)
{
   TSA *SA;
   int *x, *p;
   p=(int*)malloc((n+1)*sizeof *p);
   x=(int*)malloc((n+1)*sizeof *x);
   unsigned i;
   for (i=0; i<n; i++)
      x[i]=(int)text[i]+128;
   x[n]=0;
   suffixsort(x,p,n,256,0);
   free(x);
   unsigned k=0;
   while ((n<<k)>>(W-1) == 0) k++;
   bitsn = W-k;
   SA = new TSA();
   SA->CreatePosArray(bitsn,n);
   for (i=0; i<n; i++) {
      SA->SetPos(i,p[i+1]);
      SA->SetB(i,0);
   }
   free(p);
   return SA;
}


class TBitRank {
/* Supports rank(i)=(sum_{0<=i'<=i} B[i]) operation in constant time on bit-array B */
/* Takes less than n/(log n)+2n(loglog n)/(log n)+sqrt(n)(log n)(loglog n)+n bits */
/* That is, it takes n+o(n) bits (in practice about 1.5n bits) */   
/* Also supports operation select(x)=i, where rank(i)=x. */
/* Select is implemented making a binary search over the rank-structure, 
/* so it takes O(log(n)) time. It is possible to implement O(1)-time select,
/* using similar techniques as with rank.*/
/* See more details in David Clark's PhD thesis. */
public:   
   bool *B; // bit-array
   unsigned n; // length of B
   unsigned *firstrank; // rank of each log^2 n:th bit
   unsigned firstbits; // field length of firstrank (log n)
   unsigned firstn; // length of firstrank (n/log^2 n)/W
   unsigned *secondrank; // rank of each log n:th bit inside log^2 n size field
   unsigned secondbits; // field length of firstrank (2 log log n)
   unsigned secondn; // length of secondrank (2 n log log n/log n)/W
   unsigned *thirdrank; // ranks of all sqrt(n) fields of length (log n)/2  
   unsigned thirdbits;  // field length of thirdrank log((log n)/2)
   unsigned thirdn;  // length of thirdrank sqrt(n)(log((log n)/2)(log n)/2)/W
   unsigned *data; // stored version of B
   unsigned databits;
public:   
   TBitRank(bool *B, unsigned n) {this->B=B; this->n=n;};
   unsigned SpaceRequirementInBits() {
      return firstn*W+secondn*W+thirdn*W+n;
   }; 
   void MakeRankDictionary() {
      unsigned i,j;
      unsigned sum=0;
      /* First-level dictionary */  
      firstbits = bitsneeded(n);
      firstn = (unsigned)floor((floor(n/(1.0*firstbits))+1)*1.0/W+1)+1;
      firstrank = new unsigned[firstn];      
      j=0;
      for (i=0;i*firstbits*firstbits<n+firstbits*firstbits;i++) {
         for (;j<=i*firstbits*firstbits && j<n;j++)
            sum += B[j];
         SetField(firstrank,firstbits,i,sum); 
      }    
      /* Second-level dictionary */
      secondbits = bitsneeded(firstbits*firstbits);
      secondn = (unsigned)floor((floor(secondbits*n/(1.0*firstbits))+1)*1.0/W+1)+1+firstbits*secondbits/W+1;
      secondrank = new unsigned[secondn];
      sum = 0;
      j = 0;
      for (i=0;i*firstbits<n+firstbits*firstbits;i++) {
         if ((i*firstbits % (firstbits*firstbits))==0) {
            j = i*firstbits+1;
            sum = 0;
         }
         for (;j<=i*firstbits && j<n;j++)
            sum += B[j];
         SetField(secondrank,secondbits,i,sum); 
      }    
        /* Third-level dictionary */ 
      unsigned lognpertwo = (unsigned)floor(firstbits/2.0)+1; 
       unsigned x;
      thirdbits = bitsneeded(lognpertwo);
      unsigned thirdlength = (1u << lognpertwo); 
      thirdn = (unsigned)floor(1.0*thirdlength*lognpertwo*thirdbits/W)+1;
      thirdrank = new unsigned[thirdn];
       for (i=0;i<thirdlength;i++) {
          sum = 0;
          for (j=0;j<lognpertwo;j++) {
            if (i & (1u<<j)) sum++;
            SetField(thirdrank,thirdbits,lognpertwo*i+j,sum);    
          }              
        }         
       databits = lognpertwo;
         // convert B into unsigned[]
       data = new unsigned[n/W+1];
       for (i=0;i<n;i++) {
          if (i % W == 0) data[i/W]=0;
          data[i/W] = data[i/W] | (unsigned)B[i] << (i % W);
       }       
       //caller can delete B   
   };           
   inline unsigned rank1(unsigned i) {
      return GetField(firstrank,firstbits,i/(firstbits*firstbits));      
   };      
   inline unsigned rank2(unsigned i) {
     return GetField(secondrank,secondbits,i/firstbits);
   };
   inline unsigned rank3(unsigned i) {
      unsigned j = GetField(data,firstbits,i/firstbits);
      unsigned k = i % firstbits;
      bool isbitset = 1u & j;
      unsigned rankthree;
      if (k < databits) {
        j = (j << (W-databits)) >> (W-databits);
         rankthree = GetField(thirdrank,thirdbits,j*databits+k);
      }
      else {         
         unsigned jj = j >> databits;
         j = (j << (W-databits)) >> (W-databits);         
         rankthree = GetField(thirdrank,thirdbits,(j+1)*databits-1);
         k -= databits;
         rankthree += GetField(thirdrank,thirdbits,jj*databits+k);
      }
      return isbitset?rankthree-1:rankthree;      
   };
   inline unsigned rank(unsigned i) { // returs sum_{0<=i'<=i} B[i]
      return rank1(i)+rank2(i)+rank3(i);            
   };
   inline bool IsBitSet(unsigned i) {
      return (1u << (i % W)) & data[i/W];
   };      
   unsigned selectBasic(unsigned x) { // returns i such that x=rank(i) && rank(i-1)<x
      unsigned l=0, r=n;
      unsigned mid=(l+r)/2;      
      unsigned rankmid = rank(mid);      
      while (l+1<r) {
        if (x>rankmid) {
          l=mid;
          mid = (l+r)/2;      
        } else {
          r = mid;
          mid = (l+r)/2;    
        }     
        rankmid = rank(mid);         
      }    
      if (rankmid == x) return l;
      else return r;
   };
   inline unsigned select(unsigned x) { // returns i such that x=rank(i) && rank(i-1)<x
      // first binary search over first level rank structure
      // then binary search over second level rank structure
      // then binary search over third level rank structure
      // this is 6-7 times faster than selectBasic
      unsigned rankr;
      unsigned l=0, r=n/(firstbits*firstbits)+1;
      unsigned mid=(l+r)/2;      
      unsigned rankmid = GetField(firstrank,firstbits,mid);      
      while (l<mid) {
        if (x>rankmid)
          l = mid;
         else
          r = mid;
       mid = (l+r)/2;              
        rankmid = GetField(firstrank,firstbits,mid);         
      }    
      unsigned left;
      unsigned rankleft;
      left=l*firstbits;
      rankleft = rankmid;
          
      // now correct value is between left*firstbits+1...(left+firstbits)*firstbits
      l=0; r = firstbits;
      mid=(l+r)/2;      
      rankmid = GetField(secondrank,secondbits,left+mid);      
      while (l<mid) {
        if (x>rankleft+rankmid)
          l = mid;
        else
          r = mid;
        mid = (l+r)/2;         
        rankmid = GetField(secondrank,secondbits,left+mid);         
      } 
      left += l;
      rankleft += GetField(secondrank,secondbits,left);

      // now correct value is between left*firstbits+1...(left+1)*firstbits-1
      unsigned j = GetField(data,firstbits,left);
      unsigned jj = j >> databits; //sedondhalf
      j = (j << (W-databits)) >> (W-databits);
      rankleft -= 1u & j;      
      rankmid = GetField(thirdrank,thirdbits,(j+1)*databits-1);
      r = databits;
      if (rankleft+rankmid < x) {
         j = jj;
         left = left*firstbits+databits;
         rankleft += rankmid;
         r = firstbits-databits;
      } 
      else left = left*firstbits;      
      l=0;
      mid=(l+r)/2;      
      rankmid = GetField(thirdrank,thirdbits,j*databits+mid);      
      while (l<mid) {
        if (x>rankleft+rankmid)
          l = mid;
        else
          r = mid;
        mid = (l+r)/2;         
        rankmid = GetField(thirdrank,thirdbits,j*databits+mid);               
      }    
      if (rankleft+rankmid == x) return left+l;
      else return left+r;      
   }
   inline unsigned prev(unsigned start) {
      // returns the position of the previous 1 bit before and including start.
      // tuned to 32 bit machine
      
      unsigned i = start >> 5;
      int offset = (start % W);
      unsigned answer = start;
      unsigned int val = data[i] << (Wminusone-offset);

      if (!val) { val = data[--i]; answer -= 1+offset; }

      while (!val) { val = data[--i]; answer -= W; }

      if (!(val & 0xFFFF0000)) { val <<= 16; answer -= 16; }
      if (!(val & 0xFF000000)) { val <<= 8; answer -= 8; }

      while (!(val & 0x80000000)) { val <<= 1; answer--; }
      return answer;
      /*
      // simpler, but 3 times slower:
      while (!IsBitSet(start)) --start;      
      return start;
      */
            
   }
   void Test() {      
      unsigned i, iprev=0;
      for (i=0;i<n;i++) {
        if (rank(i)>rank(iprev)) iprev = i;
         if (rank(i)>0 && select(rank(i))!=iprev) 
            printf("select(%d)=%d!=%d\n",rank(i),select(rank(i)),iprev);                     
       }
   };   
   ~TBitRank() {
      delete firstrank;     
      delete secondrank;     
      delete thirdrank;     
      delete data;
   };      
};



/**********************************************************************************************************
 * FM-index implementation using alphabet ranking and RL-compression: RLFM-index                          *
 **********************************************************************************************************/

class TAlphabetRank {
// using fixed 0...255 alphabet
public: 
   int min; // contains ranks for symbols min...max
   int max;
   bool *B;
   TBitRank *bitrank;
   TAlphabetRank *left;
   TAlphabetRank *right;
public:
   TAlphabetRank(char *s, unsigned n, int minp, int maxp) {
      left = NULL;
      right = NULL;
      bitrank = NULL;
      B = new bool[n];
      min = minp;
      max = maxp;
      int mid = (min+max)/2;
      unsigned i, sum=0;
      for (i=0;i<n;i++) {   
         B[i] = ((int)s[i]+128)>mid;
         sum += B[i];
      }
      char *sfirst, *ssecond;
      if (n-sum > 0)  
         sfirst = new char[n-sum];
      if (sum > 0) 
         ssecond = new char[sum];
      unsigned j=0,k=0;
      for (i=0;i<n;i++)
         if (B[i]) ssecond[k++] = s[i];
         else sfirst[j++] = s[i];
      bitrank = new TBitRank(B,n);
      bitrank->MakeRankDictionary();
      //bitrank->Test();
      delete B;
      if (j > 0) {
         if (min < max) 
            left = new TAlphabetRank(sfirst,j,min,mid); 
         delete sfirst;
      }
      if (k>0) { 
         if (min < max)           
            right = new TAlphabetRank(ssecond,k,mid+1,max); 
         delete ssecond;
      }
   };   
   inline unsigned rank(int c, int i) { // returns the number of characters c before and including position i
      TAlphabetRank *temp=this;
      if (c < temp->min || c >temp->max) return 0;
      while (temp!=NULL && i>=0) {
         if (c<=(temp->min+temp->max)/2) {
            i = i-temp->bitrank->rank(i); 
            temp = temp->left; 
         }
         else { 
            i = temp->bitrank->rank(i)-1; 
            temp = temp->right;
         }
      } 
      return i+1;
   };   
   inline bool IsCharAtPos(int c, int i) {
      TAlphabetRank *temp=this;
      if (c < temp->min || c >temp->max) return false;
      while (temp!=NULL && i>=0) {
         if (c<=(temp->min+temp->max)/2) {
            if (temp->bitrank->IsBitSet(i)) return false;
            i = i-temp->bitrank->rank(i); 
            temp = temp->left; 
         }
         else { 
            if (!temp->bitrank->IsBitSet(i)) return false;         
            i = temp->bitrank->rank(i)-1; 
            temp = temp->right;
         }
      } 
      return true;
   }
   unsigned SpaceRequirementInBits() {
      unsigned bits=0;
      if (left!=NULL) bits+=left->SpaceRequirementInBits();
      if (right!=NULL) bits+=right->SpaceRequirementInBits();
      bits += bitrank->SpaceRequirementInBits();
      return bits;
   };
   bool Test(char *s, unsigned n) {
      // testing that the code works correctly
      int C[256];
      int i,j;
      bool correct=true;
      for (j=0;j<256;j++)
         C[j] = 0;
      for (i=0;i<n;i++) {
         C[(int)s[i]+128]++;
         if (C[(int)s[i]+128]!=rank((int)s[i]+128,i)) {
            correct = false;
            printf("%d (%c): %d<>%d\n",i,s[i],C[(int)s[i]+128],rank((int)s[i]+128,i)); 
         }         
      }
      return correct;            
   };   
   ~TAlphabetRank() {
      if (left!=NULL) delete left;
      if (right!=NULL) delete right;
      if (bitrank!=NULL)    
         delete bitrank;
   };   

};

void TestAlphabetRank() {
   char *s="acccbbbcccccccccaacaabaacaaaaaaacbbba";
   TAlphabetRank *ar = new TAlphabetRank(s,strlen(s),0,255);
   ar->Test(s,strlen(s));
   unsigned correct[256];
   unsigned i;
   for (i=0;i<255;i++)
      correct[i]=0;
   printf("%s\n",s);
   printf("a=%d\n",'a');
   printf("b=%d\n",'b');   
   for (i=0;i<strlen(s);i++) {
      correct[(int)s[i]+128]++;    
      printf("rank('a',%d)=%d (%d)\n",i,ar->rank((int)'a'+128,i), correct[(int)'a'+128]);
      printf("rank('b',%d)=%d (%d)\n",i,ar->rank((int)'b'+128,i), correct[(int)'b'+128]);
      printf("rank('c',%d)=%d (%d)\n",i,ar->rank((int)'c'+128,i), correct[(int)'c'+128]);      
   }
}   

char *BWT(char *text, unsigned n) {
   TSA *SA = CreateSuffixArray(text,n);
   char *s = new char[n];
   unsigned i;
   for (i=0;i<n;i++)
      if (SA->GetPos(i)==0) s[i] = text[n-1];
      else s[i] = text[SA->GetPos(i)-1];
   delete SA;
   return s;      
}

class TFMindex {
public:
    unsigned n;
    unsigned C[256];
    TAlphabetRank *alphabetrank;
public:
    TFMindex(char *text, unsigned n) {
       printf("Text size is %d B\n",n);        
       this->n = n;
       char *bwt = BWT(text,n);
       // caller can delete text now
       unsigned i, min, max;
       for (i=0;i<256;i++)
          C[i]=0;
       for (i=0;i<n;++i)
          C[(int)bwt[i]+128]++;
       for (i=0;i<256;i++)
          if (C[i]>0) {min = i; break;}          
       for (i=255;i>=min;--i)
          if (C[i]>0) {max = i; break;}                    
       unsigned prev=C[0], temp;
       C[0]=0;
       for (i=1;i<256;i++) {          
          temp = C[i];
          C[i]=C[i-1]+prev;
          prev = temp;
       }
       alphabetrank = new TAlphabetRank(bwt,n,min,max);   
       //if (alphabetrank->Test(bwt,n)) printf("alphabetrank ok\n");    
       delete bwt;
       printf("FMindex takes %d B\n",256*W/8+alphabetrank->SpaceRequirementInBits()/8);
       fflush(stdout);       
    };   
    int Search(char *pattern, unsigned m) {
       // use the FM-search replacing function Occ(c,1,i) with alphabetrank->rank(c,i)
       int c = (int)pattern[m-1]+128; unsigned i=m-1;
       int sp = C[c];
       int ep = C[c+1]-1;
       while (sp<=ep && i>=1) {
          c = (int)pattern[--i]+128;
          sp = C[c]+alphabetrank->rank(c,sp-1);
          ep = C[c]+alphabetrank->rank(c,ep)-1;
       }
       if (sp<=ep) {
          if (print) printf("%d matches found for \"%s\"\n",ep-sp+1, pattern);
          return ep-sp+1;
       } 
       else {
          if (print) printf("Pattern not found\n");
          return 0;
       }
    };   
    ~TFMindex() {
       delete alphabetrank;       
    };
};

class TRLFMindex {
public:
    unsigned n;
    unsigned len;
    unsigned C[256];
    TAlphabetRank *alphabetrank;
    TBitRank *runs_bwt;
    TBitRank *runs_sa;
public:
    TRLFMindex(char *text, unsigned n) {
       printf("Text size is %d B\n",n);        
       char *bwt = BWT(text,n);
       // caller can delete text
       // remove repeats from bwt and encode runs in a bit-vector, e.g.:
       // aaabbbbaaaacccc ==> abac, 100100010001000
       // represent the same blocks in suffix array, e.g.:
       // ==> aaa aaaa bbbb cccc ==> 100100010001000
       unsigned i, min, max, j, k;
       this->n = n;
       len = 0;
       for (i=0;i<256;i++)
         C[i]=0;
       for (i=0;i<n;++i)
          C[(int)bwt[i]+128]++;
       bool **CC = new (bool*)[256];
       for (i=0;i<256;i++)
          if (C[i]>0) 
             CC[i] = new bool[C[i]];
      	  else 
	     CC[i] = NULL;
       
       char *bwts;
       bool *B = new bool[n+1];      
       B[0] = 1; ++len;
       for (i=1;i<n;i++)
          if (bwt[i]!=bwt[i-1]) { B[i] = 1; ++len;}
	  else B[i] = 0;
       B[n] = 1;
       for (i=0;i<256;i++)
          C[i]=0;
       for (i=0;i<n;i++)
          CC[(int)bwt[i]+128][C[(int)bwt[i]+128]++] = B[i];
       bool *BB = new bool[n+1];	 	  
       k=0;
       /*for (i=0;i<256;i++)
          k += C[i];
       if (k!=n) printf("k=%d <> n=%d\n",k,n);*/
       k = 0;
       for (i=0;i<256;i++)
      	  for (j=0;j<C[i];j++)
	     BB[k++] = CC[i][j]; 
       BB[n] = 1;
       for (i=0;i<256;i++)
          if (CC[i]!= NULL) delete CC[i]; 
       delete CC;
       bwts = new char[len+1];
       j = 0;
       for (i=0;i<n;i++) 
          if (B[i]) bwts[j++]=bwt[i];
       delete bwt;	  
       /*
       for (i=1;i<len;i++)
          if (bwts[i]==bwts[i-1]) printf("errot at %d\n",i);*/
       runs_bwt = new TBitRank(B,n+1);
       runs_bwt->MakeRankDictionary();
       delete B;
       runs_sa = new TBitRank(BB,n+1);
       runs_sa->MakeRankDictionary();
       delete BB;

       // recompute C
       for (i=0;i<256;i++)
          C[i]=0;
       for (i=0;i<len;++i)
          C[(int)bwts[i]+128]++;
       for (i=0;i<256;i++)
          if (C[i]>0) {min = i; break;}          
       for (i=255;i>=min;--i)
          if (C[i]>0) {max = i; break;}                    
       unsigned prev=C[0], temp;
       C[0]=0;
       for (i=1;i<256;i++) {          
          temp = C[i];
          C[i]=C[i-1]+prev;
          prev = temp;
       }
       printf("alphabetrank next\n");
       fflush(stdout);

       alphabetrank = new TAlphabetRank(bwts,len,min,max); 
       /*for (i=0;i<len;i++)
          if (!alphabetrank->IsCharAtPos((int)bwts[i]+128,i)) printf("error at %d\n",i);  */
       //if (alphabetrank->Test(bwt,n)) printf("alphabetrank ok\n");    
       delete bwts;
       printf("RLFMindex takes %d B\n",256*W/8+alphabetrank->SpaceRequirementInBits()/8 + 
          runs_bwt->SpaceRequirementInBits()/8 + runs_sa->SpaceRequirementInBits()/8);
       fflush(stdout);       
    };   
    int Search(char *pattern, unsigned m) {
       // use the FM-search replacing function Occ(c,1,i) with alphabetrank->rank(c,i)
       int c = (int)pattern[m-1]+128; unsigned i=m-1;
       int rank;
       int sp = runs_sa->select(C[c]+1);
       int ep = runs_sa->select(C[c+1]+1)-1;
       while (sp<=ep && i>=1) {
          c = (int)pattern[--i]+128;
          rank = runs_bwt->rank(sp)-1;
          if (alphabetrank->IsCharAtPos(c,rank)) // if sp in inside the run of c
	     sp = sp-runs_bwt->prev(sp);
	  else 
	     sp = 0;
          sp += runs_sa->select(C[c]+alphabetrank->rank(c,rank-1)+1);
          rank = runs_bwt->rank(ep)-1;
          if (alphabetrank->IsCharAtPos(c,rank)) // if ep in inside the run of c
	     ep = ep-runs_bwt->prev(ep);
	  else 
	     ep = -1;
          ep += runs_sa->select(C[c]+alphabetrank->rank(c,rank-1)+1);
       }
       if (sp<=ep) {
          if (print) printf("%d matches found for \"%s\"\n",ep-sp+1, pattern);
          return ep-sp+1;
       } 
       else {
          if (print) printf("Pattern not found\n");
          return 0;
       }
    };   
    ~TRLFMindex() {
       delete alphabetrank;       
       
    };
};



